/* This is part of the netCDF package.
   Copyright 2018 University Corporation for Atmospheric Research/Unidata
   See COPYRIGHT file for conditions of use.

   Test HDF5 file code. These are not intended to be exhaustive tests,
   but they use HDF5 the same way that netCDF-4 does, so if these
   tests don't work, than netCDF-4 won't work either.

   This file creates the same file as file tst_xplatform2_3 (created
   by tst_xplatform2.c).
*/

#ifndef _H5_ERR_MACROS_H
#define _H5_ERR_MACROS_H

#define NONETCDF
#include "err_macros.h"

#endif /*_H5_ERR_MACROS_H*/
