.tabs {
 width: 100% !important;
}

.tabs ul.tablist {
 width: 100% !important;
}

.tabs ul.tablist li.linkMenu select#versions {
 vertical-align: middle;
 height: ;
 margin: 0 0 0 10px !important;
 padding: 0;
}

.tabs ul.tablist li.searchForm {
 float: right;
 padding: 0;
 margin: 0;
}

.tabs ul.tablist li.searchForm form#cse-search-box {
 float: right;
 margin: 0 10px 0 0;
}
.tabs ul.tablist li.searchForm form#cse-search-box input#search {
 border-radius: 5px;
 margin: 0 5px 0 0;
}
